/*
   socket-stream.cc

   I referred to size_io by Don Libes - National Bureau of Standards
   */


#include <sys/types.h>
#include <netinet/tcp.h>
#include "socket-stream.h"
#include <stdio.h>
#include <stdlib.h>
#include <netdb.h>
#include <sys/time.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>


extern errno;


char *SocketStream::GetServer(const char * const server_port_str, const char * const default_server)
{
  static char buf[BUFSIZE];
  ncpy_string(buf, BUFSIZE, server_port_str);
  char *c=strchr(buf, ':');
  if (c) *c='\0';
  if (!buf[0]) ncpy_string(buf, BUFSIZE, default_server);
  return buf;
}
  

US SocketStream::GetPort(const char * const server_port_str, const US default_port)
{
  char *c=strchr(server_port_str, ':');
  if (c)
  {
    c++;
    if (*c) return US(atoi(c));
  }
  return default_port;
}



SocketStream::~SocketStream(void)
{
  if (fd_socket>=0)
  {
    for (int i=0; i<MAX_FD; i++)
      if (FD_ISSET(i, &fds_socket))
	close(i);
    close(fd_socket);
  }
}


int SocketStream::Close(const int fd)
{
  if (FD_ISSET(fd, &fds_socket))
    return shutdown(fd, 2);
  else
    return close(fd);
}


int SocketStream::Init(const SocketStreamRole _role, const char *hostname, const US port, int _write_select, int sec, int usec)
{
  role=_role;
  write_select=_write_select;
  FD_ZERO(&fds);
  FD_ZERO(&fds_socket);
  now_fd=0;
  
  char buf[BUFSIZE];
  DEBUGOUTn("SockertStream:");DEBUGOUTn(hostname);DEBUGOUTn(":");DEBUGINT(port);
  if (role==SOCKET_STREAM_SERVER || !hostname || 0==strcmp(hostname, "localhost"))
  {
    if (gethostname(buf, BUFSIZE)) // ۥ̾
    {
      perror("error: SocketStream::Init(): gethostname()");
      return -1;
    }
    hostname=buf;
  }
  
  hostent *server_host;
  if (!hostname)
  {
    fprintf(stderr, "error: SocketStream::Init(): hostname is NULL\n");
    return -1;
  }
  else if (NULL==(server_host=gethostbyname(hostname))) // ۥȤΥɥ쥹
  {
    perror("error: SocketStream::Init(): gethostbyname()");
    fprintf(stderr, "\tby %s\n", hostname);
    return -1;
  }
  server_addr.Init();
  server_addr.u.in.sin_family=AF_INET;
  server_addr.u.in.sin_port=htons(port);
  memcpy((char *)&server_addr.u.in.sin_addr, server_host->h_addr, server_host->h_length);
  
  if (0>(fd_socket=socket(AF_INET, SOCK_STREAM, 0))) // åȤκ
  {
    perror("error: SocketStream::Init(): socket()");
    return -1;
  }
  /*
#ifdef DEBUG
  if (role==SOCKET_STREAM_CLIENT)
  {
    connect(fd_socket, &(server_addr.u.sa), server_addr.len);
    DEBUGOUT("successfully connected\n");
    fcntl(fd_socket, F_SETFL, 0);
    return 0;
  }
#endif
*/
  int flag=1;
  setsockopt(fd_socket, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(flag));

  if (role==SOCKET_STREAM_CLIENT)
  {
    fcntl(fd_socket, F_SETFL, O_NDELAY);
    connect(fd_socket, &(server_addr.u.sa), server_addr.len);
    struct timeval tv;
    tv.tv_sec=sec;
    tv.tv_usec=usec;
    fd_set connect_fds;
    FD_ZERO(&connect_fds);
    FD_SET(fd_socket, &connect_fds);
    int c=select(fd_socket+1, NULL, &connect_fds, NULL, &tv);
    if (c!=1)
    {
      if (c==0)
	fprintf(stderr, "error: SocketStream::Init(): connect() timmed out\n");
      else
	perror("error: SocketStream::Init(): connect()");
      close(fd_socket);
      return -1;
    }
    fcntl(fd_socket, F_SETFL, 0);
  }
  else
  {
    setsockopt(fd_socket, SOL_SOCKET, SO_REUSEADDR, (char *)0, 0);

    if (-1==bind(fd_socket, &(server_addr.u.sa), server_addr.len))
      // åȤ˥ɥ쥹ȥݡȤХ
    {
      perror("error: SocketStream::Init(): bind()");
      close(fd_socket);
      return -1;
    }
    if (listen(fd_socket, 1))
    {
      perror("error: SocketStream::Init(): listen()");
      close(fd_socket);
      return -1;
    }
  }
  // 
  return 0;
}


int SocketStream::Write(const int fd, const char *buf, const int bufsize)
{
DEBUGOUT("WRITE in");
  if (bufsize<0) return -1;
  
  UL nbufsize=htonl((UL)bufsize);
  if (sizeof(nbufsize)!=write(fd, (char *)&nbufsize, sizeof(nbufsize))) // EINTR ֤ʤ !?
  {
    close(fd);
#ifdef DEBUG
    perror("error: SocketStream::Write()");
#endif
    return -1;
  }
  
  int rbufsize=bufsize; // ĤΥХȿ
  while (rbufsize)
  {
    int cc;
    if (-1==(cc=write(fd, buf, rbufsize)))
    {
      close(fd);
#ifdef DEBUG
      perror("error: SocketStream::Write()");
#endif
      return -1;
    }
    buf+=cc;
    rbufsize-=cc;
  }
DEBUGOUT("WRITE out");
  return bufsize;
}


int SocketStream::Puts(const int fd, const char *buf)
{
  int bufsize=strlen(buf);
  int rbufsize=bufsize; // ĤΥХȿ
  while (rbufsize)
  {
    int cc;
    if (-1==(cc=write(fd, buf, rbufsize)))
    {
      close(fd);
#ifdef DEBUG
      perror("error: SocketStream::Write()");
#endif
      return -1;
    }
    buf+=cc;
    rbufsize-=cc;
  }
  return bufsize;
}


int SocketStream::is_readable(int fd, int sec, int usec)
{
  struct timeval tv;
  tv.tv_sec=sec;
  tv.tv_usec=usec;
  
  fd_set readfd;
  FD_ZERO(&readfd);
  FD_SET(fd, &readfd);
  return select(fd+1, &readfd, NULL, NULL, &tv);
}


int SocketStream::read(int fd, char* buf, int bufsize, int sec, int usec)
{
  int rbufsize=bufsize;
  
  while (rbufsize>0)
  {
    if (1!=is_readable(fd, sec, usec))
    {
      DEBUGOUT("error: SocketStream::Read():timmed out");
      return -1;
    }
    int cc;
    if (-1==(cc=::read(fd, buf, rbufsize)))
    {
#ifdef DEBUG
      perror("error: SocketStream::Read()");
#endif
      return -1;
    }
    
    if (0==cc) return 0; // EOF
    if (cc>rbufsize)
    {
      DEBUGOUT("error: SocketStream::Read(): what happened ?");
      return -1;
    }
    buf+=cc;
    rbufsize-=cc;
  }
  return bufsize;
}


int SocketStream::Read(int fd, char *buf, int bufsize, int sec, int usec)
{
  UL nbufsize;
  int cc;

  cc=read(fd, (char *)&nbufsize, sizeof(nbufsize), sec, usec);
  if (cc==-1) return -1;
  if (cc==0) // EOF
  {
    close(fd);
    return -1;
  }
  int size=ntohl(nbufsize);
  
  if (size==0) return 0;
  else if (size>bufsize)
  {
    close(fd);
    DEBUGOUT("error: SocketStream::Read(): too small bufsize");
    return -1;
  }
  
  cc=read(fd, buf, size, sec, usec);
  if (cc<=0) // EOF || Error || Timmed Out
  {
    close(fd);
    return -1;
  }
  return cc;
}


int SocketStream::Gets(int fd, char *buf, int bufsize, int sec, int usec)
{
  int i=0;
  while (i<bufsize-1)
  {
    int cc=read(fd, buf+i, 1, sec, usec);
    if (cc<=0)
    {
      close(fd);
      return -1;
    }
    i++;
    if (buf[i-1]=='\n') break;
  }
  buf[i]='\0';
  return i;
}


int SocketStream::Select(int sec, int usec)
     //  -MAX_FD-2*MAX_FD+1: 褿 write ϤǤʤ
     //                    -2: ॢ
     //                    -1: 顼
     //        0   MAX_FD-1: 褿
     //   MAX_FD 2*MAX_FD-1: 饤Ȥ³
     // 2*MAX_FD 3*MAX_FD-1: 饤Ȥ³λ
{
  struct timeval tv;
  
  FD_SET(fd_socket, &fds);
  int c;
  fd_set readfds;
  do {
    readfds=fds;
    tv.tv_sec=sec;
    tv.tv_usec=usec;
    c=select(MAX_FD, &readfds, NULL, NULL, &tv);
    if (c==0) // ॢ
      return -2;
    else if (c==-1) // 顼
    {
      if (errno==EBADF) // ¸ߤʤեǥץ
      {
	for (int i=0; i<MAX_FD; i++) // ɤ줬¸ߤʤ򸫤Ĥ
	{
	  fd_set suspect; // Ƶ
	  FD_ZERO(&suspect);
	  if (FD_ISSET(i, &fds))
	  {
	    FD_SET(i, &suspect);
	    struct timeval tv;
	    tv.tv_sec=0;
	    tv.tv_usec=0;
	    if (-1==select(MAX_FD, &suspect, NULL, NULL, &tv))
	      // ¸ߤʤեǥץ򸫤ĤΤǼ
	    {
	      FD_CLR(i, &fds);
	      FD_CLR(i, &fds_socket);
	      return i+MAX_FD*2; // 饤Ȥ³λ
	    }
	    FD_CLR(i, &suspect);
	  }
	}
      }
      else
      {
	// lets hope it was a recoverable interrupt and try again
	perror("error: SocketStream::Select()");
	return -1;
      }
    }
  } while (c==-1); //  c==-1 ȤϤʤʤϤ
  
  // given the set of ready file descriptors pick one out that is ready
  // start from where we left off, so as to give everyone service
  while (1)
  {
    now_fd=(now_fd+1)%MAX_FD;
    if (FD_ISSET(now_fd, &readfds)) break;
  }
  
  if (write_select)
  {
    if (!FD_ISSET(now_fd, &fds))
    {
      fd_set writefd;
      FD_ZERO(&writefd);
      FD_SET(now_fd, &writefd);
      tv.tv_sec=0;
      tv.tv_usec=0;
      int c=select(now_fd+1, NULL, &writefd, NULL, &tv);
      if (c!=1) return -MAX_FD-now_fd;
    }
  }
  
  if (role==SOCKET_STREAM_SERVER)
  {
    if (now_fd==fd_socket)
    {
      SockAddr from;
      int user=accept(fd_socket, &from.u.sa, &from.len);
      FD_SET(user, &fds);
      FD_SET(user, &fds_socket);

      int flag=1;
      setsockopt(user, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(flag));

      return user+MAX_FD; // 饤Ȥ³
    }
  }
  return now_fd;
}
